import copy
import os
import re
import shutil as sh

import click
from rich.box import ROUNDED
from rich.panel import Panel
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn
from rich.rule import Rule
from rich.table import Table

from ..east_context import east_command_settings
from ..helper_functions import (
    clean_up_extra_args,
    does_project_use_sysbuild,
    find_all_boards,
    find_app_build_dir,
    get_git_version,
)
from .build_type_flag import construct_extra_cmake_arguments

# This could be considered as a hack, but it is actually the cleanest way to test
# release command.
RUNNING_TESTS = False


clean_print_args = {
    "markup": False,
    "style": "",
    "overflow": "ignore",
    "crop": False,
    "highlight": False,
    "soft_wrap": False,
    "no_wrap": True,
}


def create_artefact_name(project, board, version, build_type):
    """Create an artefact name.

    Board might be in form <west_board>@<hv_version>, in that case we modify it to fit
    the artefact name.

    We also add git hash at the end if the build was not done on the clean tagged
    commit.
    """
    # As we allow project names to be paths for samples, we need to extract just the
    # name of the project.
    project = os.path.basename(project)

    board = board.replace("@", "-hv")

    # "Normalize" hw v2 board names
    board = board.replace("/", "_")

    # "release" or None build_type should not generate any build type qualifier.
    build_type = "" if build_type == "release" or not build_type else f"-{build_type}"

    git_hash = f"-{version['hash']}" if version["hash"] else ""

    return f"{project}-{board}-{version['tag']}{build_type}{git_hash}"


def create_art_dest_and_src_dir(
    name: str,
    parent: str,
    release_dir: str,
    board: str,
    build_type: str,
    is_single_app: bool,
):
    """Create destination and source directories for the artefact."""
    common_dest = os.path.join(release_dir, parent, name)

    # "Normalize" hw v2 board names
    board = board.replace("/", "_")

    if parent == "apps":
        if build_type:
            dst = os.path.join(common_dest, build_type, board)
        else:
            dst = os.path.join(common_dest, board)
        if is_single_app:
            src_dir = "app"
        else:
            src_dir = os.path.join("app", name)
    else:
        # parent is "samples"
        dst = os.path.join(common_dest, board)
        src_dir = os.path.join("samples", name)

    return dst, src_dir


def collect_binaries_using_sysbuild(build_dir, dry_run):
    """Collect binaries from build folder that used sysbuild."""
    app_build_dir = os.path.join(find_app_build_dir(build_dir), "zephyr")

    # This two always exist.
    merged_hex = os.path.join(build_dir, "merged.hex")
    zephyr_elf = os.path.join(app_build_dir, "zephyr.elf")
    binaries = [merged_hex, zephyr_elf]

    # This one might exist.
    dfu_zip = os.path.join(build_dir, "dfu_application.zip")
    if os.path.isfile(dfu_zip):
        binaries.append(dfu_zip)

    # This one also might exist, if not use zephyr.bin instead.
    app_update_bin = os.path.join(app_build_dir, "zephyr.signed.bin")
    if not os.path.isfile(app_update_bin):
        app_update_bin = os.path.join(app_build_dir, "zephyr.bin")
    binaries.append(app_update_bin)

    if dry_run or RUNNING_TESTS:
        binaries = [merged_hex, zephyr_elf, dfu_zip, app_update_bin]

    return binaries


def collect_binaries_without_using_sysbuild(build_dir, dry_run):
    """Collect binaries from build folder that did not use sysbuild."""
    app_build_dir = os.path.join(build_dir, "zephyr")

    # Determine, if we have basic build with MCUBoot, some other build with child
    # images, or default build
    if (
        os.path.isfile(os.path.join(app_build_dir, "app_update.bin"))
        or dry_run
        or RUNNING_TESTS
    ):
        # MCUBoot
        binaries = [
            "dfu_application.zip",
            "app_update.bin",
            "merged.hex",
            "zephyr.elf",
        ]
    elif os.path.isfile(os.path.join(app_build_dir, "merged.hex")):
        # Other (TFM, SPM, ...)
        binaries = ["merged.hex", "zephyr.elf"]
    else:
        # Basic build (No merged.hex is generated)
        binaries = ["zephyr.bin", "zephyr.hex", "zephyr.elf"]

    return [os.path.join(app_build_dir, binary) for binary in binaries]


def move_build_artefacts(art_name, art_dest, job_type, spdx_app_only, dry_run):
    """Moves build artefacts to art_dest and renames them to art_name."""
    os.makedirs(art_dest, exist_ok=True)

    build_dir = "build"
    # Figure out out if sysbuild is used or not, presence of domains.yaml can be used
    # for that. Determine the default build directory.
    # Collect dfu_application.zip, zephyr.signed.bin, merged.hex, zephyr.elf, first two
    # might be missing.

    if does_project_use_sysbuild(build_dir):
        binaries = collect_binaries_using_sysbuild(build_dir, dry_run)
    else:
        binaries = collect_binaries_without_using_sysbuild(build_dir, dry_run)

    for binary in binaries:
        # If doing dry run or running tests we just create empty files so that we have
        # something to copy.
        if dry_run or RUNNING_TESTS:
            path = os.path.dirname(binary)
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
            open(binary, "w").close()

        if os.path.isfile(binary):
            exten = binary.split(".")[-1]
            sh.copyfile(
                binary,
                os.path.join(art_dest, ".".join([art_name, exten])),
            )

    if spdx_app_only and job_type == "apps":
        spdx_src = os.path.join("build", "spdx")
        spdx_dest = os.path.join(art_dest, "sbom-spdx")
        sh.move(spdx_src, spdx_dest)


def show_job_summary(east, jobs):
    """Prints a summary of jobs to be run."""
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title="Jobs to run",
        box=ROUNDED,
    )
    table.add_column("Name", style="dim italic", min_width=8)
    table.add_column("Type", min_width=8)
    table.add_column("West board", style="dim italic")
    table.add_column("Build type")

    for job in jobs:
        table.add_row(
            job["src_dir"].split("/")[-1],
            "app" if job["subdir"] == "apps" else "sample",
            job["board"],
            job["build_type"] if job["subdir"] == "apps" else "/",
        )

    east.print()
    east.print(table)
    east.print()


def create_build_command(
    east,
    board=None,
    build_type=None,
    build_dir=None,
    target=None,
    source_dir=None,
    cmake_args=None,
    silence_diagnostic=False,
):
    """Helper for creating a build command.

    Returns:
        build_cmd: a string with the build command, intended to be given to run_west()
    """
    build_type_args, diagnostic = construct_extra_cmake_arguments(
        east,
        build_type,
        board,
        build_dir,
        source_dir,
    )

    if diagnostic and not silence_diagnostic:
        east.print(diagnostic)

    build_cmd = "build"

    if board:
        build_cmd += f" -b {board}"
    if build_dir:
        build_cmd += f" -d {build_dir}"
    if target:
        build_cmd += f" -t {target}"
    if source_dir:
        build_cmd += f" {source_dir}"

    # Some flags need to be passed as extra parameters to the west tool
    if build_type_args or cmake_args:
        build_cmd += " --"

    if build_type_args:
        build_cmd += f" {build_type_args}"

    if cmake_args:
        build_cmd += f" {clean_up_extra_args(cmake_args)}"

    return build_cmd


def run_job(east, progress, job, dry_run, verbose, spdx_app_only):
    """Runs the job with west.

    east ():            East context.
    job ():             Job to run.
    dry_run ():         If true then don't actually build, only show what commands
                        would run.
    verbose():          If true the Cmake output is shown.
    spdx_app_only():    If true then only apps are scanned for SPDX tags.

    Return:
        Bool            True, if job succeeded, false if it did not.
    """
    build_cmd = create_build_command(
        east,
        board=job["board"],
        build_type=job["build_type"],
        source_dir=job["src_dir"],
        silence_diagnostic=True,
    )

    if dry_run:
        east.print(
            f"[bold yellow]Dry running:[/] [dim italic]{build_cmd}[/]",
            highlight=False,
            soft_wrap=False,
            no_wrap=True,
            overflow="ignore",
            crop=False,
        )
        return

    # Handle any errors here, do not leave that to the east.run_west()
    kwargs = {"exit_on_error": False}

    if verbose:
        kwargs["silent"] = False
        kwargs["return_output"] = False
    else:
        kwargs["silent"] = True
        kwargs["return_output"] = True

    east.print(
        f"[bold green]Running:[/] [dim italic]{build_cmd}[/] ",
        highlight=False,
        soft_wrap=False,
        no_wrap=True,
        overflow="ignore",
        crop=False,
    )

    def on_failure(result):
        """Handle job failure."""
        if result["returncode"]:
            msg = (
                "Last build command [bold red]failed[/]! Check build log above"
                " to see what went wrong."
            )
            progress.stop()
            if not verbose:
                # Nicely print the build log, surounded by the rules and newlines for
                # paddings.
                east.print("")
                east.print(Rule(title="Captured build log", style="red"))
                east.print("")
                east.print(result["output"], **clean_print_args)
                east.print(Rule(title="Captured build log", style="red"))
                east.print("")
            east.print(Panel(msg, padding=1, border_style="red"))
            east.exit()

    if spdx_app_only and job["subdir"] == "apps":
        cmd = "spdx --init --build-dir build"
        result = east.run_west(cmd, **kwargs)
        on_failure(result)

    result = east.run_west(build_cmd, **kwargs)
    on_failure(result)

    if spdx_app_only and job["subdir"] == "apps":
        cmd = "spdx --build-dir build --analyze-includes --include-sdk"
        result = east.run_west(cmd, **kwargs)
        on_failure(result)


def check_if_single_app_repo():
    """Return True if this repo has just a single app project in the 'app' folder.

    Return False in any other case.
    """
    cmakelists = os.path.join("app", "CMakeLists.txt")

    # If there is no CMakeLists.txt file directly in the 'app' folder, than we are
    # certain that this is not a single app repo.
    if not os.path.isfile(cmakelists):
        return False

    # Only a "project(<something>)" is a valid indication that this is a single app repo,
    # CMakeLists without "project(<something>)" could still exist for some other
    # purpose.
    with open(cmakelists, "r") as f:
        for line in f:
            if re.match(r"project\(.*\)", line):
                return True
        return False


release_misuse_no_east_yml_msg = """
[bold yellow]east.yml[/] not found in project's root directory, [bold yellow]east release[/] needs it to determine required build steps, exiting!"""


def non_existing_app_msg_fmt(app_name):
    """Return a message."""
    return (
        f"Incorrect [bold yellow]east.yml[/], app [bold]{app_name}[/] was not"
        " found in app folder, exiting!"
    )


def non_existing_sample_msg_fmt(sample_name):
    """Return a message."""
    return (
        f"Incorrect [bold yellow]east.yml[/], sample [bold]{sample_name}[/] was not"
        " found in samples folder, exiting!"
    )


@click.command(**east_command_settings)
@click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    help=(
        "Just echo build commands and create dummy artefacts, do not actually build."
        " Dummy artefacts are placed in into [bold italic]release_dry_run[/] folder. "
    ),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Show west/Cmake build output. Default: false.",
)
@click.option(
    "--spdx-app-only",
    is_flag=True,
    help=(
        "Create an SPDX 2.2 tag-value bill of materials following the completion "
        "of a Zephyr build. This is done only for the apps, not for the samples."
    ),
)
@click.pass_obj
def release(east, dry_run, verbose, spdx_app_only):
    """Create a release folder with release artefacts.

    \b
    \n\n[bold yellow]east release[/] command runs a release process consisting of a series of [bold yellow]east build[/] commands to build applications and samples listed in the [bold yellow]east.yml[/]. Created build artefacts are then renamed and placed into [bold magenta]release[/] folder in project's root directory.

    \n\nVersion number is inferred from [bold italic cyan]git describe --tags --always --long --dirty=+[/] command. If the [bold yellow]east release[/] command is run directly on a commit with a version tag (such as [bold cyan]v1.0.2[/]), and there are no local changes, then only version tag is added to the name of artefacts, otherwise the additional git hash qualifier is added. If there is no version tag then default of [bold cyan]v0.0.0[/] is used.

    \n\nAs both [bold]apps[/] and [bold]samples[/] keys in [bold yellow]east.yml[/] are optional, release process for a specific key will be skipped, if it is not present.

    \n\nDifferent hardware versions of listed boards are picked up automatically from the `board` directory.

    \n\n[bold]Note:[/] This command requires [bold yellow]east.yml[/] to function.
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check()

    if not east.east_yml:
        east.print(release_misuse_no_east_yml_msg)
        east.exit()

    # Move into project directory, from now on all commands are run relative from this
    # location
    east.chdir(east.project_dir)

    # Clean up release folders.
    sh.rmtree("release", ignore_errors=True)
    sh.rmtree("release_dry_run", ignore_errors=True)

    # Get version from git describe command
    version = get_git_version(east)

    if dry_run:
        release_dir = "release_dry_run"
    else:
        release_dir = "release"

    # Try to get apps and samples that need to be build, if some key is not found the
    # entire release process for it is skipped.

    # Update each dict in list with parent to indicate from where it came from.
    apps = east.east_yml.get("apps", [])
    samples = east.east_yml.get("samples", [])

    # Since above two return a reference to the array, we need to make a deep copy
    # to prevent any changes to the original dict.
    apps = copy.deepcopy(apps)
    samples = copy.deepcopy(samples)

    # We inject app and samples with additional key/value pairs in below two for loops
    # so the logic afterwards for detection of jobs can be common/simpler.
    # We also do some existence checks.

    # Small adjustment for projects which only have one single app.
    listed_apps = os.listdir("app") if os.path.isdir("app") else []

    # Magical heuristic to determine if this is single app repo or not.
    is_single_app = check_if_single_app_repo()

    # cleaned_apps is here just in case if there is no app key to prevent out of bounds
    # index access when doing apps[0]
    cleaned_apps = apps[0]["name"] if len(apps) else []
    apps_in_dir = cleaned_apps if is_single_app else listed_apps

    for app in apps:
        # Check, if the app even exists before building for it
        if app["name"] not in apps_in_dir:
            east.print(non_existing_app_msg_fmt(app["name"]))
            east.exit()
        # Add parent to mark from where this key comes from
        app.update({"parent": "apps"})
        # Add "release" type (but only in apps context).
        if "build-types" in app:
            app["build-types"].append({"type": "release"})
        else:
            app["build-types"] = [{"type": None}]

    for sample in samples:
        # Check, if the sample even exists before building for it
        if not os.path.isdir(os.path.join("samples", sample["name"])):
            east.print(non_existing_sample_msg_fmt(sample["name"]))
            east.exit()
        # Add parent to mark from where this key comes from
        sample.update({"parent": "samples"})
        # Add "release" build type which for samples does nothing.
        sample.update({"build-types": [{"type": None}]})

    # Apps and samples become targets, which can be handled in similar way.
    targets = apps + samples

    jobs = []
    for target in targets:
        for board in target["west-boards"]:
            for board in find_all_boards(east, board):
                # Extract build type names from dictionaries
                build_types = [x["type"] for x in target["build-types"]]
                for build_type in build_types:
                    # Rename created binaries and move them to a specific destination.
                    # Some dance around built type name is needed for "release" type

                    dst, src_dir = create_art_dest_and_src_dir(
                        target["name"],
                        target["parent"],
                        release_dir,
                        board,
                        build_type,
                        is_single_app,
                    )

                    # Create name for job artefact
                    name = create_artefact_name(
                        target["name"], board, version, build_type
                    )

                    # Create a job dict and add it to a list
                    job = {
                        # Subdir in release folder
                        "subdir": target["parent"],
                        # Source dir, used for build command
                        "src_dir": src_dir,
                        "board": board,
                        "build_type": build_type,
                        "artefact_name": name,
                        "artefact_destination": dst,
                    }
                    jobs.append(job)

    show_job_summary(east, jobs)

    east.print(Panel("[bold]Starting jobs[/]", border_style="green"))

    progress = Progress(
        SpinnerColumn(spinner_name="shark", style="blue", speed=0.5),
        TextColumn("[bold blue]Working ... "),
        BarColumn(style="dim", complete_style="green", finished_style="green"),
        TextColumn("[bold green]{task.completed} out of {task.total} jobs done[/]"),
        console=east.console,
    )

    with progress:
        task = progress.add_task("Running build commands...", total=len(jobs))
        for job in jobs:
            # Clean old build dir
            sh.rmtree("build", ignore_errors=True)

            # Run job build
            run_job(east, progress, job, dry_run, verbose, spdx_app_only)

            # Move and rename created artefacts
            move_build_artefacts(
                job["artefact_name"],
                job["artefact_destination"],
                job["subdir"],
                spdx_app_only,
                dry_run,
            )

            # Update progress
            progress.advance(task)

    # Construct git hash for zip names
    git_hash = f"-{version['hash']}" if version["hash"] else ""

    # First try to gather all zip targets and then execute them.
    zip_targets = []

    samples_folder = os.path.join(release_dir, "samples")
    if os.path.exists(samples_folder):
        zip_targets.append(
            {
                "name": f"samples-{version['tag']}{git_hash}",
                "folder": samples_folder,
            }
        )

    for app in apps:
        # Extract build type names from dictionaries
        build_types = [x["type"] for x in app["build-types"]]

        for build_type in build_types:
            # A bit of naming manipulation for release build
            if build_type:
                app_folder = os.path.join(release_dir, "apps", app["name"], build_type)
                build_type_suf = "" if build_type == "release" else f"-{build_type}"
            else:
                app_folder = os.path.join(release_dir, "apps", app["name"])
                build_type_suf = ""

            # Build zip name and make a zip archive
            zip_targets.append(
                {
                    "name": f"{app['name']}-{version['tag']}{build_type_suf}{git_hash}",
                    "folder": app_folder,
                }
            )

    # Execute all zip targets
    for zip_target in zip_targets:
        sh.make_archive(
            os.path.join(release_dir, zip_target["name"]),
            "zip",
            zip_target["folder"],
        )

    east.print(
        Panel(
            f"[bold]Done with jobs 🎉 \n\nCheck [magenta]{release_dir}[/] folder for"
            " artefacts.[/]",
            padding=1,
            border_style="green",
        )
    )



import os

import pytest

import east.east_context

from . import helpers

# This is a conftest.py file, which has a special meaning when used with pytest.
#
# From pytest documentation:
#
# The conftest.py file serves as a means of providing fixtures for an entire directory.
# Fixtures defined in a conftest.py can be used by any test in that package without
# needing to import them (pytest will automatically discover them).
#
# IMPORTANT:
# fixtures with autouse=True are always run when tests are run.
# Other fixtures needs to explicitly mentioned in the arguments of the test.


@pytest.fixture()
def patch_constants(tmpdir_factory, monkeypatch):
    """Patch all east constants.

    This patch will move all constants under path provided by tmpdir_factory, thus
    basically creating a new environment in which we can test. That way can not override
    existing packages on the system.

    IMPORTANT: Currently unused, however it might be useful in the future.
    """
    p = tmpdir_factory.mktemp("test_dir")

    for k, v in east.east_context.const_paths.items():
        monkeypatch.setitem(east.east_context.const_paths, k, p + v)

    # Return this new path
    return p


@pytest.fixture(autouse=True)
def disable_rich_markup(monkeypatch):
    """Disable rich markup in east context init method.

    Needed in all tests, if rich markup is enabled is impossible to do comparisons
    between hard-coded rich style messages and shell's stdout.

    Args:
        monkeypatch ():     Fixture
    """
    monkeypatch.setattr(east.east_context, "RICH_CONSOLE_ENABLE_MARKUP", False)


def west_workplace_fixture_common(west_top_dir, monkeypatch, mocker):
    """Main level fixture for tests for app workspace.

    This is split so the west_top_dir dir is created separately by the callers.

    Creates a west_workplace folder on the temporary path, creates inside a set of
    folders and files expected by west and east and changes to the project directory.

    It is not expected that tests will directly use this fixture.

    Returns:
        Project path
    """
    project_path = helpers.create_good_west(west_top_dir)

    # # We pretend that
    def mocked_check_exe(self, exe):
        if exe == self.consts["nrfutil_path"]:
            return True

    def mocked_run_manager(self, command, **kwargs):
        _ = self
        _ = kwargs

        if command == "list":
            return {"output": ["v2.0.0", "v2.1.0"], "returncode": 0}

    mocker.patch("east.east_context.EastContext.check_exe", mocked_check_exe)
    mocker.patch("east.east_context.EastContext.run_manager", mocked_run_manager)

    monkeypatch.chdir(project_path)
    return project_path


@pytest.fixture()
def west_workplace(tmp_path_factory, monkeypatch, mocker):
    """Main level fixture for tests for single app workspace.

    Returns:
        Project path
    """
    west_top_dir = tmp_path_factory.mktemp("west_workplace")
    return west_workplace_fixture_common(west_top_dir, monkeypatch, mocker)


@pytest.fixture()
def west_workplace_multi_app(tmp_path_factory, monkeypatch, mocker):
    """Main level fixture for tests for multi app workspace.

    Returns:
        Project path
    """
    west_top_dir = tmp_path_factory.mktemp("west_workplace")
    project_path = west_workplace_fixture_common(west_top_dir, monkeypatch, mocker)
    helpers.create_good_west_multi_app(os.path.dirname(project_path))
    return project_path


@pytest.fixture()
def not_in_west_workplace(tmp_path_factory, monkeypatch):
    """Creates a temp workspace without anything and changes to it.

        tmp_path_factory ():
        monkeypatch ():

    Returns:
        Path
    """
    path = tmp_path_factory.mktemp("not_west_workplace")
    monkeypatch.chdir(path)
    return str(path)


@pytest.fixture(params=["single", "multi"])
def west_workplace_parametrized(tmp_path_factory, monkeypatch, mocker, request):
    """Parametrized west workspace fixture for single and multi app.

    By using it in the tests the tests are executed twice as many, first for the single
    app workspace and then for multi app workspace.

    The return dictionary contains all keys that the test should need.

    Use this fixture when behaviour under test should be the same in both single and
    multi app setup.

        tmp_path_factory ():
        monkeypatch ():
        mocker ():
        request ():

    Returns:
        Dict with project, app and prefix paths for testing.
    """
    west_top_dir = tmp_path_factory.mktemp("west_workplace")
    project_path = west_workplace_fixture_common(west_top_dir, monkeypatch, mocker)

    app_path = os.path.join(project_path, "app")
    test_path = os.path.join(project_path, "tests", "basic")
    prefix_path = "../../app/"

    if request.param == "multi":
        helpers.create_good_west_multi_app(os.path.dirname(project_path))
        app_path = os.path.join(project_path, "app", "test_one")
        prefix_path = "../../app/test_one/"

    return {
        "project": project_path,
        "app": app_path,
        "prefix": prefix_path,
        "test": test_path,
        "project_type": request.param,
    }



from .basic_commands import attach, build, bypass, clean, debug, flash, twister
from .codechecker_commands import codechecker
from .cortex_commands import cortex_debug
from .release_commands import release

__all__ = [
    "attach",
    "build",
    "bypass",
    "clean",
    "debug",
    "flash",
    "twister",
    "codechecker",
    "release",
    "cortex_debug",
]



import os
import shutil

import click
from rich.syntax import Syntax

from ..east_context import east_command_settings, east_group_settings
from ..helper_functions import clean_up_extra_args, find_app_build_dir
from .codechecker_helpers import (
    check_for_build_folder,
    check_for_codechecker,
    check_for_codechecker_config_yaml,
    check_for_compile_commands_json,
    check_for_url,
    cleanup_compile_commands_json,
    cleanup_plist_files,
    create_skip_file,
    get_endpoint,
    get_metadata_from_codecheckerfile,
)

# Print output of the parse and diff commands, with default rich highlighting, but
# without any other wrapping, cropping.
clean_print_args = {
    "highlight": True,
    "overflow": "ignore",
    "crop": False,
    "soft_wrap": False,
    "no_wrap": True,
}


@click.command(**east_command_settings)
@click.pass_obj
@click.option(
    "-h",
    "--html",
    is_flag=True,
    help="Generate a html report, instead of printing the results in the terminal."
    "Default: false.",
)
@click.option(
    "-p",
    "--dont-cleanup-plist",
    is_flag=True,
    help="Skip plist cleanup step. Default: false.",
)
@click.option(
    "-s",
    "--skip-file",
    type=str,
    help="Set skip file instead of generating one.",
)
@click.option(
    "-file",
    "--file",
    type=str,
    help="Analyze only the given file(s) not the whole compilation database. "
    "Absolute directory paths should start with '/', relative directory paths should "
    "start with '*' and it can contain path glob pattern. "
    "Example: '/path/to/main.c', 'lib/*.c', */test*'.",
)
@click.option(
    "-a",
    "--only-analyze",
    is_flag=True,
    help="Only perfrom analyze step and cleanup plist (if enabled). Default: false.",
)
@click.option(
    "-d",
    "--build-dir",
    default="build",
    type=str,
    help="Build directory to use for analysis. Default: 'build'.",
)
def check(east, html, dont_cleanup_plist, skip_file, file, only_analyze, build_dir):
    """Run [magenta bold]Codechecker[/] analysis for a built project.

    \b
    \n\nExpects that a CodeChecker config file (codechecker_config.yaml) exists in the
    project's root dir. That file is used to configure the analysis. See https://github.com/Ericsson/codechecker/tree/master/docs/config_file.md for more information. Run [bold magenta]east codechecker example-config[/] to see a suggested config file.

    \n\nResults of analysis are printed in the terminal by default. If --html flag is given, a html report is generated instead.

    \n\nBetween analysis and parsing of the results, the intermediary plist files are by default cleaned up to remove diagnostics that are not useful. To skip this step, use --dont-cleanup-plist flag.

    \n\nCurrently removed diagnostics are:

    \n\n- [bold magenta]Ineffective bitwise and operation[/], caused by [bold cyan]LOG_*[/] macros.
    \n\n- [bold magenta]Conditional operator with identical true and false expressions[/], caused by [bold cyan]SHELL_CMD_*[/], [bold cyan]APP_EVENT_MANAGER_*[/] macros.
    \n\n- [bold magenta]Value stored to 'variable' is never read or used[/], but is actually later used in disabled [bold cyan]__ASSERT*[/] or [bold cyan]LOG_*[/] macros.
    \n\n- [bold magenta]Comparison of integers of different signs: 'int' and 'unsigned int'[/], caused by [bold cyan]INIT_OBJ_RES*[/] macros.
    \n\n- [bold magenta]Missing field 'help' initializer[/], caused by [bold cyan]SHELL_SUBCMD*[/] macros.
    \n\n- [bold magenta]integer to pointer cast pessimizes optimization opportunities[/], caused by [bold cyan]LOG_*[/] macros.
    \n\n- [bold magenta]The code calls sizeof() on a pointer type. This can produce an unexpected result[/], caused by [bold cyan]LOG_*[/], [bold cyan]EVENT_LOG*[/], [bold cyan]APP_EVENT_MANAGER_LOG*[/] macros.

    \n\nBy default, a generated skip file is used to skip needless analysis of the Zephyr, NCS and external repositories. If that is not desired, you can provide your own skip file with the --skip-file flag, check the link for more info: https://codechecker.readthedocs.io/en/latest/analyzer/user_guide/#skip


    \n\n[bold]Note:[/] This command expects that the project's build folder contains compile_commands.json

    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check(check_only_west_workspace=True)

    cc = east.consts["codechecker_path"]
    cfg = os.path.join(east.project_dir, "codechecker_config.yaml")
    cc_output_dir = os.path.join(build_dir, "codechecker")
    compile_commands = os.path.join(
        find_app_build_dir(build_dir), "compile_commands.json"
    )

    check_for_codechecker(east)
    check_for_build_folder(east, build_dir)
    check_for_compile_commands_json(east, compile_commands)
    cleanup_compile_commands_json(compile_commands)
    check_for_codechecker_config_yaml(east, cfg)

    if not skip_file:
        skip_file = create_skip_file(east, build_dir, cc_output_dir)

    # Run analyze command
    analyze_cmd = (
        f"{cc} analyze --skip {skip_file} --output {cc_output_dir} "
        f"{compile_commands} --config {cfg} "
    )

    if file:
        analyze_cmd += f"--file '{file}'"

    east.run(analyze_cmd)

    if not dont_cleanup_plist:
        cleanup_plist_files(east, cc_output_dir)

    if only_analyze:
        return

    # Run parse command
    parse_cmd = f"{cc} parse --config {cfg} {cc_output_dir} "

    if html:
        parse_cmd += f"--output {cc_output_dir} --export html "

    result = east.run(parse_cmd, exit_on_error=False, return_output=True, silent=True)

    east.print(result["output"], **clean_print_args)


@click.option(
    "-a",
    "--apply",
    is_flag=True,
    help="Apply the available automatic fixes. "
    "This causes the modification of the source code. Default: false",
)
@click.option(
    "-d",
    "--build-dir",
    default="build",
    type=str,
    help="Analysed build directory. Default: 'build'.",
)
@click.command(**east_command_settings)
@click.pass_obj
def fixit(east, apply, build_dir):
    """Apply fixes suggested by the [magenta bold]Codechecker[/].

    \b
    \n\nSome analyzers may suggest some automatic bugfixes. Most of the times these are style issues which can be fixed easily. This command handles the listing and application of these automatic fixes.

    \n\n[bold]Note:[/] This command should be ran after the [bold cyan]east codechecker check[/] command.
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check(check_only_west_workspace=True)

    check_for_codechecker(east)
    cc_output_dir = os.path.join(build_dir, "codechecker")

    cc = east.consts["codechecker_path"]

    fixit_cmd = f"{cc} fixit {cc_output_dir}"

    if apply:
        fixit_cmd += " --apply"

    east.run(fixit_cmd)


example_config_text = """
analyzer:
  - --keep-gcc-include-fixed
  - --keep-gcc-intrin
  # Uncomment below option when you want to use .clang-tidy file.
  # - --analyzer-config=clang-tidy:take-config-from-directory=true
  - --analyzers
  - clang-tidy
  - clangsa
  - --enable=guideline:sei-cert
  # Enable for cross translation unit analyis, but analysis will take longer.
  # - --ctu

parse:
  - --trim-path-prefix=/*/project
  - --print-steps

store:
  - --trim-path-prefix=/*/project
"""


@click.option(
    "--url",
    type=str,
    default=lambda: os.getenv("EAST_CODECHECKER_SERVER_URL"),
    help="URL of the Codechecker server (port number is also required). "
    "If not explicitly given then value is read from the EAST_CODECHECKER_SERVER_URL "
    "env var.",
)
@click.option(
    "-d",
    "--build-dir",
    default="build",
    type=str,
    help="Analysed build directory that should be stored. Default: 'build'.",
)
@click.command(**east_command_settings)
@click.pass_obj
def store(east, url, build_dir):
    """Store the results of the [magenta bold]Codechecker[/] analysis to a server.

    \b
    \n\n[bold]Note:[/] This command should be ran after the [bold cyan]east codechecker check[/] command.
    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check(check_only_west_workspace=True)
    check_for_codechecker(east)
    check_for_url(east, url)

    cc = east.consts["codechecker_path"]
    cfg = os.path.join(east.project_dir, "codechecker_config.yaml")
    cc_output_dir = os.path.join(build_dir, "codechecker")

    name, tag = get_metadata_from_codecheckerfile(build_dir)
    endpoint = get_endpoint(east)

    store_cmd = (
        f"{cc} store --name '{name}' --url {url.strip('/')}/{endpoint} "
        f"--config {cfg} {cc_output_dir} --tag '{tag}'"
    )

    east.run(store_cmd)


@click.option(
    "-c",
    "--create",
    is_flag=True,
    help="Create example [bold magenta]codechecker_config.yaml[/] in the project's root dir.",
)
@click.command(**east_command_settings)
@click.pass_obj
def example_config(east, create):
    """Show example [bold magenta]codechecker_config.yaml[/] file."""
    east.pre_workspace_command_check(check_only_west_workspace=True)
    check_for_codechecker(east)

    if create:
        with open(os.path.join(east.project_dir, "codechecker_config.yaml"), "w") as f:
            f.write(example_config_text)

        east.print(
            "\nCreated [bold magenta]codechecker_config.yaml[/] file "
            "in project's root dir."
        )
        return

    east.print(
        "\nExample [bold magenta]codechecker_config.yaml[/] file,"
        " place this in your project's root dir:"
    )
    syntax = Syntax(
        example_config_text, "yaml", theme="ansi_dark", background_color="default"
    )
    east.print(syntax)


@click.command(
    **east_command_settings,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.option(
    "--extra-help",
    is_flag=True,
    help="Print help of the given [bold magenta]Codechecker[/] command.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED, metavar="")
def bypass(east, extra_help, args):
    """Directly run any [magenta bold]CodeChecker[/] command.

    \b
    \n\nInternally runs [magenta bold]codechecker[/] command, all given arguments are passed directly to it.

    \n\nTo learn more about possible [magenta bold]codechecker[/] arguments and options use --extra-help flag.


    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check(check_only_west_workspace=True)
    check_for_codechecker(east)

    cc = east.consts["codechecker_path"]

    cmd = f"{cc} "

    if args:
        cmd += f"{clean_up_extra_args(args)} "

    if extra_help:
        cmd += "--help"

    east.run(cmd)


@click.option(
    "--new",
    is_flag=True,
    help="Show results that don't exist in the last server analysis but exist in the "
    "local one. These are new issues that have arisen since the last store to the "
    "server.",
)
@click.option(
    "--resolved",
    is_flag=True,
    help="Show results that exist in the server analysis but aren't present in the "
    "local one. These are issues that have been resolved since the last store "
    "to the server.",
)
@click.option(
    "--unresolved",
    is_flag=True,
    help="Show results that appear both in the last server analysis and in the "
    "local analysis. These are issues that were present when last stored on the "
    "server and are still present locally.",
)
@click.option(
    "--html",
    is_flag=True,
    help="Generate a html report instead of printing the results in the terminal. "
    "Default: false.",
)
@click.option(
    "--url",
    type=str,
    default=lambda: os.getenv("EAST_CODECHECKER_SERVER_URL"),
    help="URL of the Codechecker server (port number is also required). "
    "If not explicitly given then value is read from the EAST_CODECHECKER_SERVER_URL "
    "env var.",
)
@click.option(
    "-d",
    "--build-dir",
    default="build",
    type=str,
    help="Local build directory to use for analysis. Default: 'build'.",
)
@click.command(**east_command_settings)
@click.pass_obj
def servdiff(east, new, resolved, unresolved, html, url, build_dir):
    """Compare local analysis against the last server analysis.

    \b
    \n\nUse one of the --new, --resolved, --unresolved flags to specify how to compare.

    \n\n[bold]Note:[/] This command should be ran after the [bold cyan]east codechecker check[/] command.

    \n\n[bold]Note:[/] This command can be only run from inside of a [bold yellow]West workspace[/].
    """
    east.pre_workspace_command_check(check_only_west_workspace=True)
    check_for_codechecker(east)
    check_for_url(east, url)

    cc = east.consts["codechecker_path"]
    cc_diff_output_dir = os.path.join(build_dir, "codechecker_diff")
    cc_output_dir = os.path.join(build_dir, "codechecker")

    name, _ = get_metadata_from_codecheckerfile(build_dir)
    endpoint = get_endpoint(east)

    diff_cmd = (
        f"{cc} cmd diff --basename {name} --newname {cc_output_dir} "
        f"--url {url.strip('/')}/{endpoint} "
    )

    if [new, resolved, unresolved].count(True) != 1:
        east.print(
            "\nExactly one of the [bold cyan]--new[/], [bold cyan]--resolved[/], [bold cyan]--unresolved[/] flags must be given."
        )
        east.exit()

    if new:
        diff_cmd += "--new "

    if resolved:
        diff_cmd += "--resolved "

    if unresolved:
        diff_cmd += "--unresolved "

    if html:
        shutil.rmtree(cc_diff_output_dir, ignore_errors=True)
        diff_cmd += f"--export {cc_diff_output_dir} --output html "

    result = east.run(diff_cmd, exit_on_error=False, return_output=True, silent=True)

    east.print(result["output"], **clean_print_args)

    # Propagate return code, so that if diff_cmd fails external scripts can react to
    # that.
    east.exit(result["returncode"])


@click.group(**east_group_settings, subcommand_metavar="Subcommands")
@click.pass_obj
def codechecker(east):
    """Command with several subcommands related to [magenta bold]CodeChecker[/].

    \b
    \n\nIf running CodeChecker inside continuous integration environment, run [bold cyan]export EAST_CODECHECKER_CI_MODE=1[/] before running any [bold cyan]east codechecker[/] commands.
    This will make [bold cyan]east[/] use the [magenta bold]CodeChecker[/] executable that is on the system path instead of the one in the tooling directory. The system provided [magenta bold]CodeChecker[/] will normally also want to use the system provided clang, clang-tidy and cppcheck.

    \b
    \n\nThis way users can leverage the programs provided by continuous integration
    environment and not by [bold cyan]east[/], which is usually faster due to caching.
    """
    pass


codechecker.add_command(check)
codechecker.add_command(example_config)
codechecker.add_command(fixit)
codechecker.add_command(store)
codechecker.add_command(bypass)
codechecker.add_command(servdiff)



import click
import rich_click

from .east_context import EastContext, east_group_settings
from .system_commands import connect, init, install, rtt, update
from .workspace_commands import (
    attach,
    build,
    bypass,
    clean,
    codechecker,
    cortex_debug,
    debug,
    flash,
    release,
    twister,
)

rich_click.rich_click.MAX_WIDTH = 80
rich_click.rich_click.USE_RICH_MARKUP = True
rich_click.rich_click.STYLE_HELPTEXT = "italic dim"

rich_click.rich_click.COMMAND_GROUPS = {
    "east": [
        {
            "name": "Workspace commands",
            "commands": [
                "build",
                "flash",
                "clean",
                "debug",
                "attach",
                "bypass",
                "release",
                "twister",
                "codechecker",
            ],
        },
        {
            "name": "System commands",
            "commands": ["init", "update", "install", "util"],
        },
    ],
    "east update": [{"name": "Subcommands", "commands": ["west", "env", "toolchain"]}],
    "east util": [
        {"name": "Subcommands", "commands": ["connect", "rtt", "cortex-debug"]}
    ],
    "east install": [
        {
            "name": "Packages",
            "commands": ["nrfutil-toolchain-manager", "codechecker", "toolchain"],
        }
    ],
    "east codechecker": [
        {
            "name": "Subcommands",
            "commands": [
                "check",
                "fixit",
                "example-config",
                "store",
                "servdiff",
                "bypass",
            ],
        }
    ],
}


@click.group(**east_group_settings, subcommand_metavar="Subcommands")
@click.pass_obj
def util(_):
    """Command with several subcommands related to utilities."""
    pass


util.add_command(connect)
util.add_command(rtt)
util.add_command(cortex_debug)


@click.group(
    **east_group_settings,
    chain=False,
    subcommand_metavar="<command> [command options]",
)
@click.version_option(message="%(version)s", package_name="east-tool")
@click.option(
    "--echo", is_flag=True, help="Echo each shell command before executing it."
)
@click.pass_context
def cli(ctx, echo):
    """[bold]East[/] is a command line meta-tool, useful for creating, managing and
    deploying [bold cyan]nRF Connect SDK[/] projects.

    \b
    \n\nWant to learn what each command does?

    Run [bold]east \\[command] --help[/] to show documentation for that command.

    \b
    \n\nNote that commands are split into two groups:

    - [bold]Workspace:[/] Can only be run from inside of [bold yellow]West workspace[/].

    - [bold]System:[/] Can be run from anywhere.
    """
    # EastContext object is passed to other subcommands due to the @click.pass_context
    # decorator. Additionally, the subcommands need to be decorated with @click.pass_obj
    # so they directly access the EastContext object.
    ctx.obj = EastContext(echo)
    pass


cli.add_command(build)
cli.add_command(bypass)
cli.add_command(flash)
cli.add_command(clean)
cli.add_command(init)
cli.add_command(update)
cli.add_command(install)
cli.add_command(util)
cli.add_command(release)
cli.add_command(debug)
cli.add_command(attach)
cli.add_command(twister)
cli.add_command(codechecker)


def main():
    """Main entry point for the CLI."""
    cli()


if __name__ == "__main__":
    main()



import glob
import json
import os
import platform
import shutil

from rich.panel import Panel

from ..east_context import EastContext
from ..helper_functions import download_files


def _install_nrfutil(east: EastContext, exe_path: str):
    """Install nrfutil to a proper location."""
    nrfutil = east.consts["nrfutil_path"]
    dir = os.path.dirname(nrfutil)

    shutil.rmtree(dir, ignore_errors=True)
    os.mkdir(dir)

    east.run(f"mv -f {exe_path} {nrfutil}")

    # That is octal, make it executable.
    os.chmod(nrfutil, 0o777)

    # Pin the nrfutil to a fixed version to prevent any future breaking changes.
    # Output of --version flag on linux:
    # nrfutil 7.11.1 (7c99be8 2024-05-30)
    # commit-hash: 7c99be87b691a9ea8c7d95a2190356eddad33329
    # commit-date: 2024-05-30
    # host: x86_64-unknown-linux-gnu
    # build-timestamp: 2024-05-30T12:49:04.315203564Z
    # classification: nrf-external
    east.run(f"{nrfutil} self-upgrade --to-version 7.11.1")

    # Install toolchain-manager and configure the toolchain path.
    east.run(f"{nrfutil} install toolchain-manager")

    # Below step shouldn't be done on macOS, the install-dir is hardcoded there.
    if platform.system() != "Darwin":
        east.run(
            f"{nrfutil} toolchain-manager config --set install-dir="
            f"{east.consts['east_dir']}"
        )


def _get_nrfutil_download_link():
    """Return link for the Nordic's nRF Util executable.

    The link depends on the platform that the East is running on, however it will always
    point to a lightweight launcher version of the nrfutil.

    Check https://developer.nordicsemi.com/.pc-tools/nrfutil/ for interesting stuff.
    """
    system = platform.system()

    if system == "Linux":
        system = "x64-linux"
    elif system == "Windows":
        system = "x64-windows"
    elif system == "Darwin":
        system = "universal-osx"
    else:
        print(
            f"Unsupported system ({system}), East will download x64-linux version "
            "of nrfutil."
        )
        system = "x64-linux"

    return f"https://developer.nordicsemi.com/.pc-tools/nrfutil/{system}/nrfutil"


def _get_cppcheck_download_link():
    """Just a convenience function for getting the link for the Cppcheck source code.
    This link will be updated regularly.

    """
    link = "https://github.com/danmar/cppcheck/archive/refs/tags/2.12.0.tar.gz"
    return link


def _install_cppcheck(east: EastContext, exe_path: str):
    """Install cppcheck to a proper location."""
    # FIXME: Change how you are getting path
    cppcheck_dir = os.path.join(east.consts["tooling_dir"], "cppcheck")

    # Remove old cppcheck dir if it exists
    shutil.rmtree(cppcheck_dir, ignore_errors=True)

    shutil.unpack_archive(exe_path, east.consts["tooling_dir"], format="gztar")

    # Remove version from the folder name
    dir = glob.glob(os.path.join(east.consts["tooling_dir"], "cppcheck-*"))[0]
    shutil.move(dir, cppcheck_dir)

    # Compile the cppcheck, no need to make it executable afterwards
    east.run(
        (
            f"cd {cppcheck_dir} && "
            f"make MATCHCOMPILER=yes FILESDIR={cppcheck_dir}/filesdir "
            'CXXFLAGS="-O2 -DNDEBUG -Wall -Wno-sign-compare -Wno-unused-function" -j8'
        )
    )


def _get_clang_download_link():
    """Return link for the clang+llvm binaries.

    This link will be updated regularly.
    """
    version = "16.0.0"

    arch = platform.machine().lower()

    if arch == "x86_64":
        arch = "x86_64-linux-gnu-ubuntu-18.04"
    elif arch == "aarch64":
        arch = "aarch64-linux-gnu"
    elif arch == "arm64":
        arch = "arm64-apple-darwin22.0"
    else:
        print(
            f"Unsupported architecture ({arch}), East will download x86_64 version "
            "of clang+llvm binaries."
        )
        arch = "x86_64-linux-gnu-ubuntu-18.04"

    link = (
        "https://github.com/llvm/llvm-project/releases/download/"
        f"llvmorg-{version}/clang+llvm-{version}-{arch}.tar.xz"
    )

    return link


def _install_clang_llvm(east: EastContext, exe_path: str):
    """Install clang+llvm to a proper location."""
    clang_dir = os.path.join(east.consts["tooling_dir"], "clang+llvm")

    # Remove old clang dir if it exists
    shutil.rmtree(clang_dir, ignore_errors=True)

    east.print("[bold blue]Extracing clang+llvm binaries, this will take some time...")
    shutil.unpack_archive(exe_path, east.consts["tooling_dir"], format="gztar")
    east.print("[bold green]Done!")

    # Remove version from the folder name
    dir = glob.glob(os.path.join(east.consts["tooling_dir"], "clang+llvm-*"))[0]
    shutil.move(dir, clang_dir)


def _get_codechecker_download_link():
    """Just a convenience function for getting the link for the CodeChecker source code.

    This link will be updated regularly.
    """
    version = "v6.22.2"

    link = f"https://github.com/Ericsson/codechecker/archive/refs/tags/{version}.tar.gz"

    return link


def _install_codechecker(east: EastContext, exe_path: str):
    """Install codechecker to a proper location."""
    codechecker_dir = os.path.join(east.consts["tooling_dir"], "codechecker")

    # Remove old cppcheck dir if it exists
    shutil.rmtree(codechecker_dir, ignore_errors=True)

    shutil.unpack_archive(exe_path, east.consts["tooling_dir"], format="gztar")

    # Remove version from the folder name
    dir = glob.glob(os.path.join(east.consts["tooling_dir"], "codechecker-*"))[0]
    shutil.move(dir, codechecker_dir)

    # Set extra options package so you don't need gcc-multilib and ui code.
    east.print(
        "[bold blue]Compiling codechecker from source, this will take some time..."
    )
    result = east.run(
        f"cd {codechecker_dir} && "
        "BUILD_LOGGER_64_BIT_ONLY=YES BUILD_UI_DIST=NO make package",
        exit_on_error=False,
        return_output=True,
        silent=True,
    )

    if result["returncode"] != 0:
        east.print(
            result["output"],
            markup=False,
            style="",
            overflow="ignore",
            crop=False,
            highlight=False,
            soft_wrap=False,
            no_wrap=True,
        )

        msg = (
            "Compiling codechecker [bold red]failed[/]! Check build output above"
            "to see what went wrong.\n"
            "You probably need to install some extra packages:\n\n"
            "\t[bold italic cyan]sudo apt install build-essential curl libncurses5[/]"
        )
        east.print(Panel(msg, padding=1, border_style="red"))
        east.exit()

    east.print("[bold green]Done!")

    # Set paths to the analyzers in the config file
    # Codechecker somehow needs to know which analyzers it should use, this is the only
    # way this can be set.
    config_file = os.path.join(
        codechecker_dir, "build", "CodeChecker", "config", "package_layout.json"
    )

    with open(config_file, "r") as f:
        data = json.load(f)

    data["runtime"]["analyzers"]["clangsa"] = east.consts["clang_path"]
    data["runtime"]["analyzers"]["cppcheck"] = east.consts["cppcheck_path"]
    data["runtime"]["analyzers"]["clang-tidy"] = east.consts["clang_tidy_path"]
    data["runtime"]["clang-apply-replacements"] = east.consts["clang_replace_path"]

    with open(config_file, "w") as f:
        f.write(json.dumps(data, indent=2))


toolchain_installed_msg = """
[bold magenta]nrfutil toolchain-manager[/] install done!

East will now smartly use [bold magenta]nrfutil toolchain-manager[/] whenever it can.

[bold]Note:[/] You still need to run [italic bold blue]east install toolchain[/] inside
of a [yellow bold]West workspace[/] to get the actual toolchain.
"""

cppcheck_installed_msg = """
[bold green]cppcheck install done![/]
"""

clang_llvm_installed_msg = """
[bold green]clang-tidy and clang analyzer install done![/]
"""

codechecker_installed_msg = """
[bold green]codechecker install done![/]
"""


def tool_installer(east, tool_names):
    """Install tools that are passed in the tool_names list."""
    # List of supported tools that can be installed by East.
    supported_tools = [
        {
            "name": "toolchain-manager",
            "cmd": east.consts["nrfutil_path"] + " toolchain-manager",
            "url": _get_nrfutil_download_link(),
            "install_method": _install_nrfutil,
            "installed_msg": toolchain_installed_msg,
        },
        {
            "name": "cppcheck",
            "exe": east.consts["cppcheck_path"],
            "url": _get_cppcheck_download_link(),
            "install_method": _install_cppcheck,
            "installed_msg": cppcheck_installed_msg,
        },
        {
            "name": "clang+llvm",
            "exe": east.consts["clang_path"],
            "url": _get_clang_download_link(),
            "install_method": _install_clang_llvm,
            "installed_msg": clang_llvm_installed_msg,
        },
        {
            "name": "codechecker",
            "exe": east.consts["codechecker_path"],
            "url": _get_codechecker_download_link(),
            "install_method": _install_codechecker,
            "installed_msg": codechecker_installed_msg,
        },
    ]

    tools = [tool for tool in supported_tools if tool["name"] in tool_names]

    # Construct a list of files that have to be downloaded.
    east.print("[blue]Checking for required tools...", highlight=False)

    files_to_download = []
    downloaded_files = os.listdir(east.consts["cache_dir"])

    print_args = {
        "highlight": False,
        "overflow": "ignore",
        "crop": False,
        "soft_wrap": False,
        "no_wrap": True,
    }

    for tool in tools:
        if "exe" in tool:
            tool["installed"] = east.check_exe(tool["exe"])
            tool_path = tool["exe"]
        else:
            ret = east.run(tool["cmd"], exit_on_error=False, silent=True)
            tool["installed"] = ret["returncode"] == 0
            tool_path = tool["cmd"]

        if tool["installed"]:
            east.print(f"{tool_path} [green]found", **print_args)
        elif tool["name"] in downloaded_files:
            east.print(
                f"{tool_path} [red]not installed[/], but downloaded file is "
                f"present in the {east.consts['cache_dir']}",
                **print_args,
            )
        else:
            east.print(f"{tool_path} [red]not found", **print_args)
            files_to_download.append(
                {
                    "url": tool["url"],
                    "name": tool["name"],
                }
            )

    if all([tool["installed"] is True for tool in tools]):
        east.print("\n[green]All required tools are installed.")

    # Download all required files, which are actually programs or installer scripts.
    download_files(files_to_download, east.consts["cache_dir"])

    # Run an installation method for packages that are not installed.
    for tool in tools:
        if not tool["installed"]:
            exe_path = os.path.join(east.consts["cache_dir"], tool["name"])
            tool["install_method"](east, exe_path)

    # WARN: We assume that download_files will never fail. Depending on the urgency we
    # should implement better handling.
    for tool in tools:
        if not tool["installed"]:
            east.console.rule("", style="")
            east.print(tool["installed_msg"])
    east.console.rule("", style="")



import copy
import json
import os

import click
import rich.prompt

from ..east_context import east_command_settings
from ..helper_functions import determine_svd_file, get_cortex_debug_params


@click.command(**east_command_settings)
@click.pass_obj
@click.option(
    "-d",
    "--build-dir",
    type=str,
    default="build",
    help=("Build directory of the project to debug. Default: build."),
)
@click.option(
    "-r",
    "--rtt-config",
    is_flag=True,
    help=(
        "Add RTT configuration to the launch.json file. With it the Cortex Debug "
        "starts an RTT client in the VSCode. If set then any external RTT clients "
        "won't be able to connect to the GDB server created by the Cortex "
        "Debug. Default: False."
    ),
)
@click.option(
    "--device",
    type=str,
    help=(
        "Set the target device directly, instead of getting it from the "
        "[bold]runners.yaml[/] file."
    ),
)
@click.option(
    "--gdb-path",
    type=str,
    help=(
        "Set the GDB path directly, instead of getting it in the "
        "[bold]runners.yaml[/] file."
    ),
)
@click.option(
    "--elf-file",
    type=str,
    help=("Set the elf file directly. Default [bold]build/zephyr/zephyr.elf[/]."),
)
@click.option(
    "--svd-file",
    type=str,
    help=("Set the SVD file directly, instead of trying to determine it."),
)
@click.option(
    "--no-py",
    is_flag=True,
    help=(
        "Use regular arm-zepher-eabi-gdb instead of python version for gdbPath in "
        "config. Enable this if you are facing an error while trying to use RTT "
        "in VSCode. Default: False."
    ),
)
def cortex_debug(
    east, build_dir, rtt_config, device, gdb_path, elf_file, svd_file, no_py
):
    """Create a configuration file for the [bold green]Cortex Debug[/] VScode extension.

    \b
    \n\nRunning this command inside the application project directory creates a [bold].vscode/launch.json[/] file that configures the [bold green]Cortex Debug[/] for debugging that project.

    \b
    \n\nIf the command is run without any extra options then it will try to infer the required info from the [bold]runners.yaml[/] file. If that is not desired then the --device, --gdb and --elf options can be used to provide info directly.

    \b
    \n\nSVD files for Nordic chips are determined automatically. You can override this with the --svd-file option.

    """
    east.pre_workspace_command_check()

    # If not all params are provided, try to infer them from the runners.yaml file.
    if not all([device, gdb_path, elf_file]):
        try:
            _device, _gdb_path, _elf_file = get_cortex_debug_params(build_dir)
            device = device if device else _device
            gdb_path = gdb_path if gdb_path else _gdb_path
            elf_file = elf_file if elf_file else _elf_file
        except Exception as e:
            east.print(f"Error: {e}. Can't create the .vscode/launch.json file.")
            east.exit()

    if no_py:
        gdb_path = gdb_path.replace("-py", "")

    attach = {
        "name": "Attach",
        "device": device,
        "executable": elf_file,
        "cwd": "${workspaceFolder}",
        "request": "attach",
        "type": "cortex-debug",
        "runToEntryPoint": "main",
        "servertype": "jlink",
        "gdbPath": gdb_path,
    }

    if rtt_config:
        attach["rttConfig"] = {
            "enabled": True,
            "address": "auto",
            "decoders": [{"port": 0, "type": "console"}],
        }

    if not svd_file:
        svd_file = determine_svd_file(east, device)

    if svd_file:
        attach["svdFile"] = svd_file

    # Debug JSON object is mostly the same.
    debug = copy.deepcopy(attach)
    debug["name"] = "Debug"
    debug["request"] = "launch"

    launch = {
        "version": "0.2.0",
        "configurations": [debug, attach],
    }

    vscode_dir = os.path.join(east.project_dir, ".vscode")
    os.makedirs(vscode_dir, exist_ok=True)

    launch_file = os.path.join(vscode_dir, "launch.json")
    if os.path.isfile(launch_file):
        if not rich.prompt.Confirm.ask(
            "File [bold magenta].vscode/launch.json[/] already exists. "
            "Do you want to overwrite it (y/n)?",
            show_choices=False,
            show_default=False,
        ):
            east.exit()

    with open(launch_file, "w") as f:
        json.dump(launch, f, indent=4)

    east.print(
        "Created [bold magenta].vscode/launch.json[/] file in project directory."
    )



from .basic_west_commands import init, update
from .install import install
from .jlink_commands import connect, rtt

__all__ = [
    "init",
    "update",
    "install",
    "util",
    "rtt",
    "connect",
]



import os
from typing import List

from ..helper_functions import return_dict_on_match


def is_child_in_parent(parent, child):
    """Check if the child path is in parent path
        parent (): path
        child ():  path
    Returns: True if yes, otherwise False.
    """
    # Smooth out relative path names, note: if you are concerned about symbolic
    # links, you should use os.path.realpath too
    parent = os.path.abspath(parent)
    child = os.path.abspath(child)

    # Compare the common path of the parent and child path with the common path of
    # just the parent path. Using the commonpath method on just the parent path will
    # regularise the path name in the same way as the comparison that deals with
    # both paths, removing any trailing path separator
    return os.path.commonpath([parent]) == os.path.commonpath([parent, child])


def _construct_required_cmake_args(
    conf_files: List[str], board: str, path_prefix: str, source_dir: str
) -> str:
    """Constructs required cmake args from conf_files. Adds board specific conf file if it
    is found.
    Path prefix is applied, for cases where sample inherits a list conf_files.

        conf_files (List(str)):     List of conf files
        board (str):                West board name
        path_prefix (str):          Path prefix added to conf files

    Returns:
        Cmake args
    """
    # Determine if there is some prefix involved
    prefix = f"{path_prefix}/" if path_prefix else ""
    # We always use common.conf
    cmake_args = f"-DCONF_FILE={prefix}conf/common.conf"

    overlay_configs = []

    # If a west board is given search for board specific config file and add it to the
    # start of overlay_configs, if it is found.
    if board:
        # Sanitize the board input, user might gave hv version
        board = board.split("@")[0]
        # Cleanup also hw v2 board names
        board = board.replace("/", "_")

        # Location of the board file depends on the source_dir and prefix
        board_prefix = f"{source_dir}/{prefix}" if source_dir else f"{prefix}"
        board_conf = f"{board_prefix}conf/{board}.conf"

        if os.path.isfile(board_conf):
            overlay_configs.append(f"{board}.conf")

    # Then add conf_files if there are any
    overlay_configs += conf_files

    # Glue together configs
    if len(overlay_configs) > 0:
        overlay_config = ";".join([f"{prefix}conf/{file}" for file in overlay_configs])
        cmake_args += f' -DOVERLAY_CONFIG="{overlay_config}"'

    return f"{cmake_args}"


def _check_previous_build(east, build_dir, build_type_str):
    """Checks if rebuild is needed by looking at the contents of the
    last_build_type_flag file.

    Args:
        east ():            East context
        build_dir ():       Location of build_dir
        build_type_str ():  Build type in string format

    Returns:
        one of the following strings:
        - "nobuild" if there is no build directory.
        - "rebuild" if rebuild is needed.
        - "just_build" if rebuild is not needed.
    """
    build_dir = build_dir.strip("/") if build_dir else "build"
    check_file = os.path.join(build_dir, "last_build_type_flag")

    if not os.path.isdir(build_dir):
        os.mkdir(build_dir)
        with open(check_file, "w") as f:
            f.write(build_type_str)
        return "nobuild"

    # This one handles a edge case where user might have deleted the check_file
    if not os.path.isfile(check_file):
        with open(check_file, "w") as f:
            f.write(build_type_str)
        return "rebuild"

    with open(check_file, "r") as f:
        last_build_type = f.read()

    with open(check_file, "w") as f:
        f.write(build_type_str)

    if last_build_type == build_type_str:
        return "just_build"
    else:
        return "rebuild"


build_type_misuse_no_east_yml_msg = """
[bold yellow]east.yml[/] not found in project's root directory, option [bold
cyan]-/u--build-type[/] needs it to determine [bold]Kconfig[/] overlay files, exiting!"""

build_type_misuse_msg = """
Option [bold cyan]--build-type[/] can only be used inside of the application folder, exiting!"""

build_type_misuse_no_app_msg = """
Option [bold cyan]--build-type[/] can only be used when apps key in [bold yellow]east.yml[/] has at least one
application entry!"""


def no_build_type_msg(build_type):
    """Returns a message that a given build type does not exist in east.yml."""
    return (
        f"\nGiven --build-type [bold]{build_type}[/] does [bold red]not"
        " exist[/] for this app!"
    )


def construct_extra_cmake_arguments(east, build_type, board, build_dir, source_dir):
    """Construct extra cmake arguments for west build command.

    This function will construct cmake_arguments (specifically values for OVERLAY_CONFIG
    and CONF_FILE defines), based on:
    * given build type
    * given board
    * contents of east.yml file
    * contents of build folder if found

    The function behaviour is described:
    * in docs/configuration.md document - That document basically describes how should
    east build behave.
    * in tests/test_build_type.py - Tests the behaviour east build command in a
    variety of different scenarios.

        east ():            East context
        build_type ():      Build type given from the east build
        board ():           Board given from the east build
        build_dir ():       Location of build_dir, if None then "./build" is used

    Returns:  Tuple of strings
                First string are extra cmake arguments for west build command should be
                placed after `--` in west command
                Second string is diagnostic message that can be printed. It might not be
                given so it should be checked first.
    """
    if not east.east_yml:
        if not build_type:
            # east.yml is optional, if it is not present present then default to plain
            # west behaviour: no cmake args
            return ("", "")
        else:
            east.print(build_type_misuse_no_east_yml_msg)
            east.exit()

    # Set to empty string, it will stay like that unless we are inside app that uses
    # build types
    cmake_build_type = ""

    # Modify current working dir, if source_dir is used, rstrip is needed cause
    # path.join adds one "/" if joining empty string.
    source_dir = source_dir if source_dir else ""
    cwd = os.path.join(east.cwd, source_dir).rstrip("/")

    # Does the relative path contain app or samples string
    relpath = os.path.relpath(cwd, start=east.project_dir)
    inside_app = "app" in relpath
    inside_sample = "samples" in relpath

    if not inside_app:
        if build_type:
            # --build-type can only be given from inside of the project dir
            # (--source-dir can be given to point to some app), or directly inside app.
            # Any other location is not allowed.
            east.print(build_type_misuse_msg)
            east.exit()
        if not build_type and not inside_sample:
            # We are not inside app and not inside sample, no build type was given, we
            # are default to plain west behaviour: no cmake args.
            return ("", "")

    # We get past this point if we are inside app or sample.

    # apps field is optional, using build type without that field is however not
    # allowed.
    app_array = east.east_yml.get("apps")
    if build_type and not app_array:
        east.print(build_type_misuse_no_app_msg)
        east.exit()

    # If inside samples determine in which sample are we, get its element
    if inside_sample:
        # samples key is optional, if it is not present in east_yml then we default to
        # plain west behaviour: no cmake args
        sample_array = east.east_yml.get("samples")
        if not sample_array:
            return ("", "")

        sample_name = os.path.basename(cwd)
        sample_dict = return_dict_on_match(sample_array, "name", sample_name)

        if not sample_dict or "inherit-build-type" not in sample_dict:
            # In case where there is no inherit, or sample is not listed we default to
            # plain west behavior: no cmake args.
            return ("", "")

        # Determine if we have to inherit from some app or we can just use prj.conf
        inherited_app = sample_dict["inherit-build-type"]["app"]
        build_type = sample_dict["inherit-build-type"]["build-type"]
        app = return_dict_on_match(app_array, "name", inherited_app)
        path_to_project_dir = os.path.relpath(east.project_dir, start=cwd)

        if len(app_array) == 1:
            path_prefix = os.path.join(path_to_project_dir, "app")
        else:
            path_prefix = os.path.join(path_to_project_dir, "app", inherited_app)

    build_types_str = build_type if build_type else "release"

    if inside_app:
        # If we do not have an app array and we there is no build type, we default to
        # plaing west behaviour: no cmake args.
        if not app_array:
            return ("", "")

        # Determine what kind of project it is, single or multi app
        if os.path.basename(cwd) == "app":
            app = app_array[0]
        else:
            # Get the folder name that we are in
            app = return_dict_on_match(app_array, "name", os.path.basename(cwd))

            if not app:
                # We are inside app folder that is not listed in east.yaml, we default
                # to plain west behaviour: no cmake args.
                return ("", "")
        path_prefix = ""
        cmake_build_type = f' -DEAST_BUILD_TYPE="{build_types_str}"'

    if "build-types" not in app:
        if build_type:
            east.print(no_build_type_msg(build_type))
            east.exit()
        else:
            # If no --build-type is given and there is not build-types key in the app,
            # we default to plain west behaviour: no cmake args.
            return ("", "")

    # "release" is a special, implicit, default, build type. Samples can request to
    # inherit from it, in that case only the common.conf is added to the build.
    if build_type == "release":
        build_type = None

    # Extract a list of conf files from the app, exit if they do not exist.
    conf_files = []
    if build_type:
        matched_type = return_dict_on_match(app["build-types"], "type", build_type)
        if not matched_type:
            east.print(no_build_type_msg(build_type))
            east.exit()

        conf_files = matched_type["conf-files"]

    # Construct required cmake args,
    required_cmake_args = _construct_required_cmake_args(
        conf_files, board, path_prefix, source_dir
    )

    # Check build type of previous build
    result = _check_previous_build(east, build_dir, build_types_str)

    if result == "just_build":
        # Just build, no need to construct previous cmake args
        return ("", "")

    if result == "rebuild":
        msg = (
            "[italic bold dim]💬 Old settings found in the build folder, deleting "
            "it and rebuilding[/]"
        )
        east.run(f"rm -rf {build_dir}")
    elif result == "nobuild":
        # Previous cmake args are empty string, no build folder was found
        msg = "[italic bold dim]💬 Build folder not found, running CMake build[/]"
    return required_cmake_args + cmake_build_type, msg



import os
import shutil as sh

import east
from east.east_context import EastContext
from east.helper_functions import find_all_boards

from . import helpers
from .helpers import helper_test_against_west_run


def test_finding_hardware_versions(west_workplace):
    """Test finding hardware versions."""
    _ = west_workplace
    east = EastContext()

    # Some board that has as several different versions in board folder
    boards = find_all_boards(east, "custom_nrf52840dk")
    expected_boards = [
        "custom_nrf52840dk@1.0.0",
        "custom_nrf52840dk@1.1.0",
        "custom_nrf52840dk@2.20.1",
    ]
    assert expected_boards == boards

    # Some board that has a board folder but not revisions
    boards = find_all_boards(east, "nrf52840dk/nrf52840")
    expected_boards = ["nrf52840dk/nrf52840"]
    assert expected_boards == boards


def test_no_east_yml_with_release(west_workplace_parametrized, monkeypatch, mocker):
    """Running east release with not east.yml should abort."""
    os.remove("east.yml")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "release",
        should_succed=False,
    )


east_yaml_no_samples_key = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: uart
        conf-files:
          - debug.conf
          - uart.conf
"""

expected_app_release_west_commands = [
    (
        "build -b custom_nrf52840dk@1.0.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf"'
        ' -DEAST_BUILD_TYPE="debug"'
    ),
    (
        "build -b custom_nrf52840dk@1.0.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
        ' -DEAST_BUILD_TYPE="uart"'
    ),
    (
        "build -b custom_nrf52840dk@1.0.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DEAST_BUILD_TYPE="release"'
    ),
    (
        "build -b custom_nrf52840dk@1.1.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf"'
        ' -DEAST_BUILD_TYPE="debug"'
    ),
    (
        "build -b custom_nrf52840dk@1.1.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
        ' -DEAST_BUILD_TYPE="uart"'
    ),
    (
        "build -b custom_nrf52840dk@1.1.0 app -- -DCONF_FILE=conf/common.conf"
        ' -DEAST_BUILD_TYPE="release"'
    ),
    (
        "build -b custom_nrf52840dk@2.20.1 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf"'
        ' -DEAST_BUILD_TYPE="debug"'
    ),
    (
        "build -b custom_nrf52840dk@2.20.1 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
        ' -DEAST_BUILD_TYPE="uart"'
    ),
    (
        "build -b custom_nrf52840dk@2.20.1 app -- -DCONF_FILE=conf/common.conf"
        ' -DEAST_BUILD_TYPE="release"'
    ),
    (
        "build -b nrf52840dk_nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;conf/debug.conf"'
        ' -DEAST_BUILD_TYPE="debug"'
    ),
    (
        "build -b nrf52840dk_nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;conf/debug.conf;conf/uart.conf"'
        ' -DEAST_BUILD_TYPE="uart"'
    ),
    (
        "build -b nrf52840dk_nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
        ' -DEAST_BUILD_TYPE="release"'
    ),
]


def test_basic_app_release_behaviour(west_workplace, monkeypatch, mocker):
    """Running east release with no samples key, should skip build process for samples and
    build apps.
    """
    project_path = west_workplace

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_no_samples_key,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "release",
        expected_west_cmds=expected_app_release_west_commands,
    )


def test_basic_app_release_behaviour_no_samples_folder(
    west_workplace, monkeypatch, mocker
):
    """Running east release with no samples key and no samples folder should skip build
    process for samples and build apps.
    """
    project = west_workplace

    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_no_samples_key,
    )

    sh.rmtree(os.path.join(project, "samples"))

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project,
        "release",
        expected_west_cmds=expected_app_release_west_commands,
    )


east_yaml_no_apps_key = """
samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
"""

expected_samples_release_west_commands = [
    "build -b custom_nrf52840dk@1.0.0 samples/settings",
    "build -b custom_nrf52840dk@1.1.0 samples/settings",
    "build -b custom_nrf52840dk@2.20.1 samples/settings",
    "build -b nrf52840dk_nrf52840 samples/settings",
    "build -b custom_nrf52840dk@1.0.0 samples/dfu",
    "build -b custom_nrf52840dk@1.1.0 samples/dfu",
    "build -b custom_nrf52840dk@2.20.1 samples/dfu",
    "build -b nrf52840dk_nrf52840 samples/dfu",
]


def test_basic_samples_release_behaviour(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Running east release with no apps key, should skip build process for apps and build
    samples.
    """
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "release",
        expected_west_cmds=expected_samples_release_west_commands,
    )


def test_basic_release_behaviour(west_workplace, monkeypatch, mocker):
    """Running east release with both apps and samples keys should run normally."""
    project_path = west_workplace

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_no_apps_key + east_yaml_no_samples_key,
    )
    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "release",
        expected_west_cmds=expected_app_release_west_commands
        + expected_samples_release_west_commands,
    )


east_yaml_no_build_types_key = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
"""

expected_app_release_west_commands_no_build_types = [
    "build -b custom_nrf52840dk@1.0.0 app",
    "build -b custom_nrf52840dk@1.1.0 app",
    "build -b custom_nrf52840dk@2.20.1 app",
    "build -b nrf52840dk_nrf52840 app",
]


def test_basic_app_release_behaviour_no_build_type(west_workplace, monkeypatch, mocker):
    """Running east release on applications without any build types should just build
    apps.
    """
    project_path = west_workplace

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_no_build_types_key,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "release",
        expected_west_cmds=expected_app_release_west_commands_no_build_types,
    )


east_yaml_samples_no_build_types_key = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
"""

expected_app_samples_release_west_commands = [
    "build -b custom_nrf52840dk@1.0.0 app",
    "build -b custom_nrf52840dk@1.1.0 app",
    "build -b custom_nrf52840dk@2.20.1 app",
    "build -b nrf52840dk_nrf52840 app",
    "build -b custom_nrf52840dk@1.0.0 samples/settings",
    "build -b custom_nrf52840dk@1.1.0 samples/settings",
    "build -b custom_nrf52840dk@2.20.1 samples/settings",
    "build -b nrf52840dk_nrf52840 samples/settings",
    "build -b custom_nrf52840dk@1.0.0 samples/dfu",
    "build -b custom_nrf52840dk@1.1.0 samples/dfu",
    "build -b custom_nrf52840dk@2.20.1 samples/dfu",
    "build -b nrf52840dk_nrf52840 samples/dfu",
]


def test_basic_app_release_behaviour_no_build_type_with_samples(
    west_workplace, monkeypatch, mocker
):
    """Running east release on applications and samples without any build types should
    just build apps and samples.
    """
    project_path = west_workplace

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_samples_no_build_types_key,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "release",
        expected_west_cmds=expected_app_samples_release_west_commands,
    )


east_yaml_non_existing_sample = """
samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

  - name: non_existing_sample_name
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
"""


def test_east_yml_with_non_existant_samples(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Running east release when east.yml contains non-existing samples should abort before
    issuing any of the west build commands.
    """
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_non_existing_sample,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "release",
        should_succed=False,
    )


east_yaml_non_existing_app = """
apps:
  - name: test_one
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_two
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_three_which_does_not_exist
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
"""


def test_east_yml_with_non_existant_apps(west_workplace_multi_app, monkeypatch, mocker):
    """Running east release when east.yml contains non-existing apps should abort before
    issuing any of the west build commands. This only make sense in multi app
    workspaces.
    """
    project = west_workplace_multi_app

    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_non_existing_app,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project,
        "release",
        should_succed=False,
    )


east_yaml_single_app_simple = """
apps:
  - name: test_one
    west-boards:
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
"""

expected_single_app_release_west_commands = [
    (
        "build -b nrf52840dk_nrf52840 app/test_one -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;conf/debug.conf"'
        ' -DEAST_BUILD_TYPE="debug"'
    ),
    (
        "build -b nrf52840dk_nrf52840 app/test_one -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
        ' -DEAST_BUILD_TYPE="release"'
    ),
]


def test_building_app_that_is_not_on_the_first_level(
    west_workplace_multi_app, monkeypatch, mocker
):
    """Usecase: you have an `app` folder, with two projects inside it. `app_one`,
    is listed in east.yml, `app_two` is not. East should be able to build `app_one`
    without a problem.
    """
    project = west_workplace_multi_app

    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_single_app_simple,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project,
        "release",
        expected_west_cmds=expected_single_app_release_west_commands,
    )


east_yaml_hw_model_v2_only_apps = """
apps:
  - name: test_one
    west-boards:
      - custom/nrf52840dk
      - nrf52840dk/nrf52840
"""

expected_hw_v2_only_apps_release_west_commands = [
    "build -b custom/nrf52840dk@1.0.0 app",
    "build -b custom/nrf52840dk@1.1.0 app",
    "build -b custom/nrf52840dk@2.20.1 app",
    "build -b nrf52840dk/nrf52840 app",
]


def test_hw_model_v2_only_apps(west_workplace, monkeypatch, mocker):
    """Running east release with only apps key, where the board names are in the new
    hardware model v2 format.
    """
    helpers.create_and_write(
        west_workplace,
        "east.yml",
        east_yaml_hw_model_v2_only_apps,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace,
        "release",
        expected_west_cmds=expected_hw_v2_only_apps_release_west_commands,
    )


east_yaml_hw_model_v2_with_build_types = """
apps:
  - name: test_one
    west-boards:
      - nrf52840dk/nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
"""

expected_hw_v2_with_build_types_release_west_commands = [
    (
        "build -b nrf52840dk/nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;conf/debug.conf"'
        ' -DEAST_BUILD_TYPE="debug"'
    ),
    (
        "build -b nrf52840dk/nrf52840 app -- -DCONF_FILE=conf/common.conf"
        ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
        ' -DEAST_BUILD_TYPE="release"'
    ),
]


def test_hw_model_v2_with_build_types(west_workplace, monkeypatch, mocker):
    """Running east release with build types, where the board names are in the new
    hardware model v2 format, should work.
    """
    helpers.create_and_write(
        west_workplace,
        "east.yml",
        east_yaml_hw_model_v2_with_build_types,
    )

    monkeypatch.setattr(east.workspace_commands.release_commands, "RUNNING_TESTS", True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace,
        "release",
        expected_west_cmds=expected_hw_v2_with_build_types_release_west_commands,
    )



import os

import pytest
from click.testing import CliRunner

from east.__main__ import cli

from . import helpers
from .helpers import helper_test_against_west_run


def test_build_type_with_no_east_yml(west_workplace_parametrized, monkeypatch):
    """--build-type option can not be used if east.yml is not found in the project root.
    East needs to exit with error and message.
    """
    os.remove("east.yml")

    monkeypatch.chdir(west_workplace_parametrized["app"])
    runner = CliRunner()

    east_cmd = "build --build-type debug".split(" ")

    # Call from the project root
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1


# if in project but not in app or samples then --build-type should not be given,
# we are assuming that user might be trying to run build on tests or zephyr's samples or
# something else
def test_build_type_outside_project_dir(west_workplace, monkeypatch):
    """No --build-type should be given if east is called:
    - inside of project dir, but not in app
    - outside of project dir
    East needs to exit with error and message.
    """
    project_path = west_workplace
    runner = CliRunner()

    east_cmd = "build --build-type debug".split(" ")

    # Call from the project root
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1

    # Go one level up from project dir
    monkeypatch.chdir(os.path.dirname(project_path))
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1

    # Go into zephyr folder
    monkeypatch.chdir(os.path.join(os.path.dirname(project_path), "zephyr"))
    result = runner.invoke(cli, east_cmd)
    assert result.exit_code == 1


# in single app folder behaviour
@pytest.mark.parametrize(
    "build_type_flag, cmake_arg",
    [
        ("", '-DCONF_FILE=conf/common.conf -DEAST_BUILD_TYPE="release"'),
        (
            "--build-type debug",
            '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/debug.conf" -DEAST_BUILD_TYPE="debug"',
        ),
        (
            "--build-type uart",
            '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf" -DEAST_BUILD_TYPE="uart"',
        ),
    ],
)
def test_build_type_single_app_behaviour(
    west_workplace, monkeypatch, mocker, build_type_flag, cmake_arg
):
    """Build command needs to parse the --build-type flag into appopriate command line
    arguments.

    This test case is for a single app behaviour.
    """
    project_path = west_workplace
    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(project_path, "app"),
        f"build {build_type_flag}",
        expected_west_cmd=f"build -- {cmake_arg}",
    )


@pytest.mark.parametrize(
    "multiapp, build_type_flag, cmake_arg",
    [
        ("test_one", "", '-DCONF_FILE=conf/common.conf -DEAST_BUILD_TYPE="release"'),
        (
            "test_one",
            "--build-type debug",
            (
                '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/debug.conf"'
                ' -DEAST_BUILD_TYPE="debug"'
            ),
        ),
        (
            "test_one",
            "--build-type uart",
            (
                "-DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf;conf/uart.conf"'
                ' -DEAST_BUILD_TYPE="uart"'
            ),
        ),
        ("test_two", "", '-DCONF_FILE=conf/common.conf -DEAST_BUILD_TYPE="release"'),
        (
            "test_two",
            "--build-type rtt",
            (
                '-DCONF_FILE=conf/common.conf -DOVERLAY_CONFIG="conf/rtt.conf"'
                ' -DEAST_BUILD_TYPE="rtt"'
            ),
        ),
        (
            "test_two",
            "--build-type debug-rtt",
            (
                "-DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf;conf/rtt.conf"'
                ' -DEAST_BUILD_TYPE="debug-rtt"'
            ),
        ),
    ],
)
def test_build_type_multi_app_behaviour(
    west_workplace_multi_app, monkeypatch, mocker, multiapp, build_type_flag, cmake_arg
):
    """Build command needs to parse the --build-type flag into appopriate command line
    arguments.

    This test case is for a multi app behaviour.
    """
    project_path = west_workplace_multi_app

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(project_path, "app", multiapp),
        f"build {build_type_flag}",
        expected_west_cmd=f"build -- {cmake_arg}",
    )


def create_last_build_type_file(app_path, build_type, build_dir="build"):
    """Create last_build_type_file file inside build folder."""
    build_type_str = build_type.split(" ")[-1]
    if build_type_str == "":
        build_type_str = "release"

    helpers.create_and_write(
        app_path, f"{build_dir}/last_build_type_flag", build_type_str
    )


@pytest.mark.parametrize(
    "build_type_flag",
    [
        "",
        "--build-type debug",
        "--build-type uart",
    ],
)
def test_build_type_build_folder_behaviour_same_flags(
    west_workplace_parametrized, monkeypatch, mocker, build_type_flag
):
    """If the build folder with same conf files with that --build-type expects exits then
    no cmake args are added to the build command to avoid cmake rebuilds.
    """
    app_path = west_workplace_parametrized["app"]
    create_last_build_type_file(app_path, build_type_flag)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        app_path,
        f"build {build_type_flag}",
        expected_west_cmd="build",
    )


@pytest.mark.parametrize(
    "build_type_flag, overlay_configs",
    [
        (
            "--build-type debug",
            "conf/debug.conf",
        ),
        (
            "--build-type uart",
            "conf/debug.conf;conf/uart.conf",
        ),
    ],
)
def test_build_type_build_folder_behaviour_different_flags(
    west_workplace_parametrized, monkeypatch, mocker, build_type_flag, overlay_configs
):
    """If the build folder exists but it has build flags that are not expected by the east
    then rebuild is triggered.
    """
    app_path = west_workplace_parametrized["app"]

    create_last_build_type_file(app_path, "--build-type release")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        app_path,
        f"build {build_type_flag}",
        expected_west_cmd=(
            "build -- -DCONF_FILE=conf/common.conf"
            f' -DOVERLAY_CONFIG="{overlay_configs}" '
            f'-DEAST_BUILD_TYPE="{build_type_flag.split(" ")[1]}"'
        ),
    )


def test_build_type_non_existant_type(west_workplace_parametrized, monkeypatch, mocker):
    """If given --build-type does not exists then east needs to exit and throw message."""
    app_path = west_workplace_parametrized["app"]

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        app_path,
        "build --build-type asasdada",
        should_succed=False,
    )


def test_build_type_samples_with_build_type_option(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Running east build with --build-type inside samples should fail."""
    project_path = west_workplace_parametrized["project"]
    sample_path = os.path.join(project_path, "samples", "settings")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build --build-type debug",
        should_succed=False,
    )


def test_build_type_samples_inherit(west_workplace_parametrized, monkeypatch, mocker):
    """Test inherit key normal use."""
    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return (
            f"build -- -DCONF_FILE={prefix}conf/common.conf"
            f' -DOVERLAY_CONFIG="{prefix}conf/debug.conf"'
        )

    sample_path = os.path.join(project_path, "samples", "settings")
    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


def test_build_type_samples_inherit_build_folder_same_flag(
    west_workplace_parametrized, monkeypatch, mocker
):
    """In case where sample (with an inherit keword) has a existing build folder from
    before, no extra cmake args should be emmited.
    """
    project_path = west_workplace_parametrized["project"]

    sample_path = os.path.join(project_path, "samples", "settings")

    create_last_build_type_file(sample_path, "--build-type debug")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd="build",
        should_succed=True,
    )


def test_build_type_samples_no_inherit(
    west_workplace_parametrized, monkeypatch, mocker
):
    """In case where there is no inherit we default to basic west behavior: no cmake args."""
    project_path = west_workplace_parametrized["project"]
    sample_path = os.path.join(project_path, "samples", "dfu")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd="build",
        should_succed=True,
    )


def test_build_type_samples_does_not_exist(
    west_workplace_parametrized, monkeypatch, mocker
):
    """In case where sample does not exist in east.yml we default to basic west behaviour:
    no cmake args.

    """
    project_path = west_workplace_parametrized["project"]

    sample_path = os.path.join(project_path, "samples", "super_duper_sample")
    os.mkdir(sample_path)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd="build",
        should_succed=True,
    )


east_yaml_non_existing_inheriting = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_two
        build-type: debug
"""


def test_non_existing_inherited_app(west_workplace_parametrized, monkeypatch, mocker):
    """In case where sample is inheriting from a non-existing app we exit."""
    project_path = west_workplace_parametrized["project"]

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yaml_non_existing_inheriting,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "build",
        should_succed=False,
    )


east_yaml_inheriting_from_release_single_app = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: release
"""

east_yaml_inheriting_from_release_multi_app = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_two
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: release
"""


def test_inheriting_from_release_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """It should be possible for a sample to inherit from a "release" build type. In that
    case only common.conf should be added to the build.
    """
    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return f"build -- -DCONF_FILE={prefix}conf/common.conf"

    if west_workplace_parametrized["project_type"] == "single":
        east_yml = east_yaml_inheriting_from_release_single_app

    if west_workplace_parametrized["project_type"] == "multi":
        east_yml = east_yaml_inheriting_from_release_multi_app

    helpers.create_and_write(
        project_path,
        "east.yml",
        east_yml,
    )

    sample_path = os.path.join(project_path, "samples", "settings")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


east_yaml_duplicated_app_names = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""

east_yaml_duplicated_samples_names = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: settings
    west-boards:
      - custom_nrf52840dk

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""

east_yaml_duplicated_build_types = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
      - type: debug
        conf-files:
          - debug.conf
      - type: debug
        conf-files:
          - debug.conf
      - type: rtt
        conf-files:
          - debug.conf

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug

  - name: dfu
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840
    # Don't inherit, use prj.conf in the sample's folder.
"""


@pytest.mark.parametrize(
    "east_yml",
    [
        "east_yaml_duplicated_app_names",
        "east_yaml_duplicated_build_types",
        "east_yaml_duplicated_samples_names",
    ],
)
def test_duplicated_names_in_east_yml(
    west_workplace_multi_app, monkeypatch, mocker, east_yml
):
    """In case where east.yml has duplicated app names exit and throw error."""
    helpers.create_and_write(
        west_workplace_multi_app,
        "east.yml",
        east_yml,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_multi_app,
        "build",
        should_succed=False,
    )


@pytest.mark.parametrize(
    "east_cmd, expected_west_cmd",
    [
        (
            "build -b nrf52840dk_nrf52840",
            (
                "build -b nrf52840dk_nrf52840 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
                ' -DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840 --build-type uart",
            (
                "build -b nrf52840dk_nrf52840 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
                ' -DEAST_BUILD_TYPE="uart"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840@1.0.0",
            (
                "build -b nrf52840dk_nrf52840@1.0.0 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
                ' -DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840@1.0.0 --build-type uart",
            (
                "build -b nrf52840dk_nrf52840@1.0.0 -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
                ' -DEAST_BUILD_TYPE="uart"'
            ),
        ),
        (
            "build -b nonexisting_board",
            "build -b nonexisting_board -- -DCONF_FILE=conf/common.conf"
            ' -DEAST_BUILD_TYPE="release"',
        ),
        (
            "build -b nonexisting_board --build-type debug",
            (
                "build -b nonexisting_board -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
                ' -DEAST_BUILD_TYPE="debug"'
            ),
        ),
    ],
)
def test_searching_for_west_board_specific_confs(
    west_workplace_parametrized,
    monkeypatch,
    mocker,
    east_cmd,
    expected_west_cmd,
):
    """In case where we are building for a specific board, build command needs to check in
    the conf folder if a file with name west_board.conf exists, if yes then it needs to
    add it.
    In sample folders with inherit key it needs to do the same thing.
    """
    project_path = west_workplace_parametrized["app"]

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        east_cmd,
        expected_west_cmd,
        should_succed=True,
    )


def test_different_build_dir_path_empty_dir(
    west_workplace_parametrized, monkeypatch, mocker
):
    """With different build dir path, but no build folder the expected west command
    should just include the -d option and common.conf.
    """
    project_path = west_workplace_parametrized["app"]

    build_dir = "../different_build_dir"

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        f"build -d {build_dir}",
        f"build -d {build_dir} -- -DCONF_FILE=conf/common.conf "
        '-DEAST_BUILD_TYPE="release"',
        should_succed=True,
    )


def test_different_build_dir_path_full_dir_same_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """With different build dir path and build folder that has same previous build type
    files as current ones the west command should just include -d option,
    and nothing else.
    """
    project_path = west_workplace_parametrized["app"]

    build_dir = "../different_build_dir"
    build_type = "--build-type uart"

    create_last_build_type_file(project_path, build_type, build_dir=build_dir)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        f"build -d {build_dir} {build_type}",
        f"build -d {build_dir}",
        should_succed=True,
    )


def test_different_build_dir_path_full_dir_different_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """With different build dir path and build folder that has different previous build
    type files as current ones the west command should include -d option and current
    conf files.
    """
    project_path = west_workplace_parametrized["app"]

    build_dir = "../different_build_dir"

    old_build_type = "--build-type uart"
    new_build_type = "--build-type debug"
    new_overlay_configs = "conf/debug.conf"

    create_last_build_type_file(
        project_path, build_type=old_build_type, build_dir=build_dir
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        f"build -d {build_dir} {new_build_type}",
        f"build -d {build_dir} -- -DCONF_FILE=conf/common.conf"
        f' -DOVERLAY_CONFIG="{new_overlay_configs}" -DEAST_BUILD_TYPE="debug"',
        should_succed=True,
    )


@pytest.mark.parametrize(
    "east_cmd, expected_west_cmd",
    [
        (
            "build app/test_one",
            (
                "build app/test_one -- -DCONF_FILE=conf/common.conf "
                '-DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build app/test_two",
            (
                "build app/test_two -- -DCONF_FILE=conf/common.conf "
                '-DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build app/test_one --build-type debug",
            (
                "build app/test_one -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
                ' -DEAST_BUILD_TYPE="debug"'
            ),
        ),
        (
            "build app/test_two --build-type debug",
            (
                "build app/test_two -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
                ' -DEAST_BUILD_TYPE="debug"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840 app/test_one",
            (
                "build -b nrf52840dk_nrf52840 app/test_one --"
                " -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
                ' -DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840 --build-type uart app/test_one",
            (
                "build -b nrf52840dk_nrf52840 app/test_one --"
                " -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
                ' -DEAST_BUILD_TYPE="uart"'
            ),
        ),
    ],
)
def test_using_different_source_dirs(
    west_workplace_multi_app,
    monkeypatch,
    mocker,
    east_cmd,
    expected_west_cmd,
):
    """Using a different source dir must not affect the extra args after `--`."""
    project_path = west_workplace_multi_app

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        east_cmd,
        expected_west_cmd,
        should_succed=True,
    )


def test_sample_with_inherit_and_with_source_dir(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Test if sample with inherit key and source_dir works."""
    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return (
            f"build samples/settings -- -DCONF_FILE={prefix}conf/common.conf"
            f' -DOVERLAY_CONFIG="{prefix}conf/debug.conf"'
        )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "build samples/settings",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


def test_sample_with_inherit_and_with_source_dir_and_board(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Test if sample with inherit key and source_dir works, when there is board involved."""
    project_path = west_workplace_parametrized["project"]

    def west_cmd_fmt(prefix):
        return (
            "build -b nrf52840dk_nrf52840 samples/settings --"
            f" -DCONF_FILE={prefix}conf/common.conf"
            " -DOVERLAY_CONFIG="
            f'"{prefix}conf/nrf52840dk_nrf52840.conf;{prefix}conf/debug.conf"'
        )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project_path,
        "build -b nrf52840dk_nrf52840 samples/settings",
        expected_west_cmd=west_cmd_fmt(west_workplace_parametrized["prefix"]),
        should_succed=True,
    )


east_yaml_no_apps_key = """
samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
"""


def test_no_apps_key_in_east_yml_build_type(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Apps key is optional.
    Running east build with no apps key in east.yml should fail if --build-type is
    given.
    """
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build --build-type debug",
        should_succed=False,
    )


def test_no_apps_key_in_east_yml_app(west_workplace_parametrized, monkeypatch, mocker):
    """Apps key is optional.
    Running east build with no apps key in east.yml and building for app should not emit
    any extra cmake args.
    """
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build",
        "build",
    )


def test_no_apps_key_in_east_yml_sample(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Apps key is optional.
    Running east build with no apps key in east.yml should not emit any extra cmake
    args.
    """
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(west_workplace_parametrized["project"], "samples", "settings"),
        "build",
        "build",
        should_succed=True,
    )


east_yaml_no_samples_key = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
      - nrf52840dk_nrf52840

    build-types:
      - type: debug
        conf-files:
          - debug.conf
"""


def test_no_samples_key_in_east_yml(west_workplace_parametrized, monkeypatch, mocker):
    """Samples key is optional.
    Running east build with no samples key in east.yml in samples should not emit any
    extra args.
    """
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_no_samples_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        os.path.join(west_workplace_parametrized["project"], "samples", "settings"),
        "build",
        "build",
        should_succed=True,
    )


east_yaml_empty_apps_key = """
apps:
"""


def test_empty_apps_key(west_workplace_parametrized, monkeypatch, mocker):
    """Empty apps key is not allowed."""
    helpers.create_and_write(
        west_workplace_parametrized["project"],
        "east.yml",
        east_yaml_empty_apps_key,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build",
        should_succed=False,
    )


@pytest.mark.parametrize(
    "build_path",
    [
        "project",
        "test",
    ],
)
def test_building_outside_of_app_and_samples(
    west_workplace_parametrized, build_path, monkeypatch, mocker
):
    """Empty apps key is not allowed, and samples can not inherit from it."""
    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized[build_path],
        "build -b native_posix",
        "build -b native_posix",
        should_succed=True,
    )


def test_building_app_that_is_not_in_east_yaml(
    west_workplace_parametrized, monkeypatch, mocker
):
    """East shouldn't add extra flags to the west build call when building an app that
    isn't in the east.yaml.
    """
    unlisted_app = os.path.join(
        west_workplace_parametrized["project"], "app/test_three"
    )
    os.makedirs(unlisted_app, exist_ok=True)

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["project"],
        "build -b native_posix app/test_three",
        "build -b native_posix app/test_three",
        should_succed=True,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        unlisted_app,
        "build -b native_posix",
        "build -b native_posix",
        should_succed=True,
    )


def test_empty_east_yml_is_valid(west_workplace_parametrized, monkeypatch, mocker):
    """Apps and sample keys are both optional in east.yaml, so empty east.yml is valid."""
    # Create empty east.yml
    open(os.path.join(west_workplace_parametrized["project"], "east.yml"), "w").close()

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build -b native_posix",
        "build -b native_posix",
        should_succed=True,
    )


east_yaml_no_build_type = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk
"""


def test_no_build_type_is_ok(west_workplace_parametrized, monkeypatch, mocker):
    """East should not add any extra flags to the west build call when no build type is
    specified in the east.yaml.
    """
    project = west_workplace_parametrized["project"]
    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_no_build_type,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build -b custom_nrf52840dk",
        expected_west_cmd="build -b custom_nrf52840dk",
    )


def test_build_type_as_arg_is_not_ok_if_not_in_east_yml(
    west_workplace_parametrized, monkeypatch, mocker
):
    """East should abort if build type is given as an argument but built-types key is
    not present in the east.yaml.
    """
    project = west_workplace_parametrized["project"]
    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_no_build_type,
    )

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        west_workplace_parametrized["app"],
        "build -b custom_nrf52840dk --build-type debug",
        should_succed=False,
    )


east_yaml_no_build_type_inheriting = """
apps:
  - name: test_one
    west-boards:
      - custom_nrf52840dk

samples:
  - name: settings
    west-boards:
      - custom_nrf52840dk
    inherit-build-type:
        app: test_one
        build-type: debug
"""


def test_inheriting_from_an_app_without_build_type_should_fail(
    west_workplace_parametrized, monkeypatch, mocker
):
    """Test that inheriting from an app without build type fails."""
    project = west_workplace_parametrized["project"]
    helpers.create_and_write(
        project,
        "east.yml",
        east_yaml_no_build_type_inheriting,
    )

    sample_path = os.path.join(project, "samples", "settings")

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        sample_path,
        "build -b custom_nrf52840dk",
        should_succed=False,
    )


@pytest.mark.parametrize(
    "east_cmd, expected_west_cmd",
    [
        (
            "build --snippet rtt app/test_one",
            (
                "build --snippet rtt app/test_one -- -DCONF_FILE=conf/common.conf "
                '-DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build --snippet rtt app/test_two",
            (
                "build --snippet rtt app/test_two -- -DCONF_FILE=conf/common.conf "
                '-DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build --snippet rtt app/test_one --build-type debug",
            (
                "build --snippet rtt app/test_one -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
                ' -DEAST_BUILD_TYPE="debug"'
            ),
        ),
        (
            "build --snippet rtt app/test_two --build-type debug",
            (
                "build --snippet rtt app/test_two -- -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/debug.conf"'
                ' -DEAST_BUILD_TYPE="debug"'
            ),
        ),
        (
            "build -b nrf52840dk_nrf52840 --snippet rtt app/test_one",
            (
                "build -b nrf52840dk_nrf52840 --snippet rtt app/test_one --"
                " -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf"'
                ' -DEAST_BUILD_TYPE="release"'
            ),
        ),
        (
            "build --snippet rtt -b nrf52840dk_nrf52840 --build-type uart app/test_one",
            (
                "build --snippet rtt -b nrf52840dk_nrf52840 app/test_one --"
                " -DCONF_FILE=conf/common.conf"
                ' -DOVERLAY_CONFIG="conf/nrf52840dk_nrf52840.conf;'
                'conf/debug.conf;conf/uart.conf"'
                ' -DEAST_BUILD_TYPE="uart"'
            ),
        ),
    ],
)
def test_additional_flags_should_passthrough(
    west_workplace_multi_app, monkeypatch, mocker, east_cmd, expected_west_cmd
):
    """Test that additional flags, such as --snippet, pass through east logic unchanged."""
    project = west_workplace_multi_app

    helper_test_against_west_run(
        monkeypatch,
        mocker,
        project,
        east_cmd,
        expected_west_cmd,
    )
